package edu.columbia.sip;

import java.net.ServerSocket;
import java.net.Socket;
import java.net.InetAddress;
import java.io.IOException;

/** The SIPServer is the main entry point class in the SIP package.  
 * It runs as a program and behaves either as a listening server (proxy or 
 * redirecting) or as a User Agent.  This class uses the other classes in the
 * package to implement the standard defined in {put RFC here}/Internet Draft XXX
 *
 * @author Moshe Sambol
 */
public class SIPServer extends Thread {
    /** the default port SIP servers listen on, as defined in the draft.*/
    public final static int DEFAULT_PORT = 5060;
    
    /** these "enums" are used to indicate in which mode the server is running. */
    public final static int FORWARD_SERVER = 1;
    public final static int PROXY_SERVER = 2;
    public final static int USER_AGENT = 3;
    
    /** holds the mode of the server - must be one of the above enums. */
    private int _mode;
    
    /** the actual port this server is listening on (and writing to, 
     * for a user agent.*/
    private int _port;
    
    /** the main socket for communication.  For servers, additional sockets
     * will be spawned off from this one as connections are made.  User Agents 
     * send requests out on this socket.
     */
    private ServerSocket _mainsock;
    
    /** a SIP URL which represents the return address of this SIP server */
    private SIPURL _myurl;
    
    /** the registration manager is responsable for keeping track of SIPServers
     * and user agents which are registered with this server and assorted other
     * lists (eg. the list of AvailabilityListeners, etc.)
     */
    private static RegistrationManager _registrationManager;
    
    /** the callID manager is responsable for keeping track of calls which 
     * are currently active and in which this server is participating.
     */
    private static CallIDManager _callIDManager;
    
    /** the location service is used to find other SIPServers and to find users
     * on those servers.  */
    private static LocationService _locationService;
    
    /** the entry point for when this class is run as a program.  It accepts
     * several parameters:  the first tells it whether to run as a proxy, forwarding,
     * or user agent, and the rest provide additional optional information including
     * the port to use for primary connections if the user would like to specify 
     * a non-default value. <br>
     * if no parameters are passed or "-help" is passed in, the program 
     * prints usage information.     
     */
    public static void main(String args[]) {
        int mode=0, port=0;
        
        // check the command line parameters, print usage info if necessary.
        if (args.length == 0) {
            usage();
        } else {
            if (args[0].equals("-proxy")) {
                mode = PROXY_SERVER;
            } else if (args[0].equals("-forward")) {
                mode = FORWARD_SERVER;
            } else if (args[0].equals("-useragent")) {
                mode = USER_AGENT;
            } else {
                usage();
            }
      
            if (args.length == 1) {
                // only a mode was provided
                port = DEFAULT_PORT;
                if (mode == USER_AGENT) {
                    System.out.println("Warning: you have specified that a user " +
                    "agent be run on the default port, usually reserved for SIP " +
                    "servers.  If there is a SIP server running on this machine " +
                    "this will cause a conflict.  To run an agent on another port " +
                    "use the parameter -port=<port>\n");
                }                   
            } else {
                // check additional arguments provided in key=value format                
                String arg=null, key=null, value=null;
                for (int i=1; i<args.length; i++) {
                    arg = args[i];
                    int equalPos = arg.indexOf('=');
                    if (equalPos == -1) {
                        // an '=' was not found in the argument
                        System.out.println("Badly formatted argument " + arg);
                        usage();
                    }
                    key = arg.substring(0, equalPos);
                    value = arg.substring(equalPos + 1, arg.length());
                    System.out.println("DEBUG: key = '" + key + "', value = '" + 
                                        value + "'");
                      
                    // "switch" of all legal key-value pairs                                        
                    if (key.equals("-port")) {  
                        port = Integer.parseInt(value);
                    } else {
                        System.out.println("Badly formatted argument " + arg);
                        usage();
                    }
                } // done with key-value pair checking
            }
        } // done checking command line parameters

        try {
            SIPServer server = new SIPServer(mode, port);
            server.start();
        } catch (IOException ioe) {
            System.out.println("SIPServer cannot initialize communication: " + ioe);
            System.exit(1);
        }
    } // end main()
    
    
    /** Constructor.
     *
     * @param mode the mode of the server: one of the "enums" defined above.
     * @param port the port to run on.  '0' indicates 'any'.
     *
     * @throws IOException if there is an error setting up the main socket
     */
    public SIPServer(int mode, int port) throws IOException {
        _mode = mode;
        _port = port;
        
        _registrationManager = new RegistrationManager();
        _callIDManager = new CallIDManager();
        
        // change this line to use other location services (perhaps based on
        // a command line switch) when they are available.
        _locationService = new DefaultLocationService();
        
        // set up the socket which we will listen on
        try {
            _mainsock = new ServerSocket(port);
            System.out.println("SIPServer(): main ServerSocket listening on port " + _mainsock.getLocalPort());
        } catch (IOException ioe) {
            throw new IOException("Error creating main socket: " + ioe);
        }

        _myurl = new SIPURL("sip:" + getAddress());
    } // end CTOR(mode, port)
    
    public void run() {
//        if (mode != USER_AGENT) {
            // loop forever waiting for connection
            Socket childsock = null;
            SIPConnection connection = null;
            while (true) {
                try {
                    childsock = _mainsock.accept();
                    connection = new SIPConnection(childsock);
                    connection.start();
                    } catch (IOException ioe) {
                        System.out.println("SIPServer error handling incoming " +
                                            "connection: " + ioe);
                    }
            }
  //      }
    } // end run()


    /** prints usage information and exits. */
    public static void usage() {
        System.out.println("Usage: java edu.columbia.sip.SIPServer <mode> <params>");
        System.out.println("Modes:");
        System.out.println("    -forward");
        System.out.println("    -help");
        System.out.println("    -proxy");
        System.out.println("    -useragent");
        System.out.println();  
        System.out.println("Parameters:");
        System.out.println("    -port=<portnum>");
        System.out.println("\n");
        System.exit(1);
    } // end usage()

    /** this method is used to tell the server that another entity is interested
     * in availability information which the server may have.  It is called in one of
     * two ways:  1) If a HERE request is recieved, a RemoteAvailabilityListener is 
     * added to the list to represent the remote entity.  2) If this server is a user agent, 
     * this method may be called directly as a convenience method for actually
     * becoming an availability listener on whatever server is appropriate for the subject.
     * For that reason, the method actually does two things: adds the listener to its own
     * internal list, and uses the location service to see if there is another apropriate
     * server with information about the subject.  If there is, this server adds itself
     * as an availability listener for the subject. <p>
     * If notification comes in as a result
     * of this subscription and there is no longer any local party interested in the information,
     * the server will remove its subscription from the remote server. <p>
     * To reduce the occurence of miscommunications the server also adds the callID of the
     * outgoing request to the list of open callIDs which it maintains.  
     */
    public void addAvailabilityListener(AvailabilityListener al) {
        if (al == null) {
            System.out.println("SIPServer.addAvailabilityListener(): ERROR: al is null.");
            return;
        }
        
        _registrationManager.addAvailabilityListener(al);

        // get the identity being requested in the al
        SIPURL subject = al.getURL();
        if (subject == null) {
            System.out.println("SIPServer.addAvailabilityListener(): ERROR: al's subject is null.");
            return;
        }
        
        // find its location - with which SIP server does it register
        SIPURL loc = _locationService.lookup(subject);
        
        if (loc != null) {
            // send a HERE request to that location, where this server should be treated
            //   as an availability listener.
            

            try {
               SIPRequest hereReq = new SIPRequest(SIPToken.HERE, subject, getURL());
               hereReq.setHeader(SIPToken.TO, loc);
               hereReq.respondToServer(this);
 //               System.out.println("SIPServer.aal(): creating connection to the server...");
                SIPConnection conn = new SIPConnection(loc);
 //               System.out.println("SIPServer.aal(): established connection, sending message.");
                conn.putMessage(hereReq);
                conn.close();
            } catch (Exception e) {
                System.out.println("SIPServer.addAvailabilityListener(): error sending HERE for " +
                                    subject + "to SIP server " + loc + ": " + e);
            }
        }
    } // end addAvailabilityListener()
    
    public void removeAvailabilityListener(AvailabilityListener al) {
        _registrationManager.removeAvailabilityListener(al);
        // then remove them from the other place(s) where we added them as 
        // remote listeners by checking with the call-id manager.
        // remove them there by sending a cancel with that call-id.
        
    }
    
    /** get the registration manager for this process */
    public static RegistrationManager getRegistrationManager() {
        return _registrationManager;
    }
    
    /** get the call ID manager for this process */
    public static CallIDManager getCallIDManager() {
        return _callIDManager;
    }
    
    /** 
     * get the TCP port on which this server is listening.
     */
    public int getTCPPort() {
        return _mainsock.getLocalPort();
    }
    
    public static String getAddress() {
        try {
        //    return InetAddress.getLocalHost().getHostName();
            return InetAddress.getLocalHost().getHostAddress();
        } catch (java.net.UnknownHostException uhe) {
            return null;
        }
    }
    
    /** return a SIP URL representing the return address of this SIP server */
    public SIPURL getURL() {
        return _myurl;
    }
    
    public String toString() {
        return (getAddress() + String.valueOf(_port));
    }
    
    /** this method should be called when the SIPServer is about to be shut down. 
     * it will take care of cleaning up state and ending calls-in-progress gracefully. 
     */
    public void shutDown() {
        ///...
    }
    
} // end class SIPServer